# searchAgents.py
# ---------------
# Licensing Information:  You are free to use or extend these projects for 
# educational purposes provided that (1) you do not distribute or publish 
# solutions, (2) you retain this notice, and (3) you provide clear 
# attribution to UC Berkeley, including a link to 
# http://inst.eecs.berkeley.edu/~cs188/pacman/pacman.html
# 
# Attribution Information: The Pacman AI projects were developed at UC Berkeley.
# The core projects and autograders were primarily created by John DeNero 
# (denero@cs.berkeley.edu) and Dan Klein (klein@cs.berkeley.edu).
# Student side autograding was added by Brad Miller, Nick Hay, and 
# Pieter Abbeel (pabbeel@cs.berkeley.edu).


"""
This file contains all of the agents that can be selected to
control Pacman.  To select an agent, use the '-p' option
when running pacman.py.  Arguments can be passed to your agent
using '-a'.  For example, to load a SearchAgent that uses
depth first search (dfs), run the following command:

> python pacman.py -p SearchAgent -a fn=depthFirstSearch

Commands to invoke other search strategies can be found in the
project description.

Good luck and happy searching!
"""
from game import Directions
from game import Agent
from game import Actions
import util
import time
import copy

class SearchProblem:
    """
    This class outlines the structure of a search problem, but doesn't implement
    any of the methods (in object-oriented terminology: an abstract class).

    You do not need to change anything in this class, ever.
    """

    def getStartState(self):
        """
        Returns the start state for the search problem
        """
        util.raiseNotDefined()

    def isGoalState(self, state):
        """
          state: Search state

        Returns True if and only if the state is a valid goal state
        """
        util.raiseNotDefined()

    def getSuccessors(self, state):
        """
          state: Search state

        For a given state, this should return a list of triples,
        (successor, action, stepCost), where 'successor' is a
        successor to the current state, 'action' is the action
        required to get there, and 'stepCost' is the incremental
        cost of expanding to that successor
        """
        util.raiseNotDefined()

    def getCostOfActions(self, actions):
        """
         actions: A list of actions to take from the initial state

        This method returns the total cost of a particular sequence 
        of actions, executed from the initial state.
        The sequence must be composed of legal moves
        """
        util.raiseNotDefined()


class PositionSearchProblem(SearchProblem):
    """
    A search problem defines the state space, start state, goal test,
    successor function and cost function.  This search problem can be
    used to find paths to a particular point on the pacman board.

    The state space consists of (x,y) positions in a pacman game.

    Note: this search problem is fully specified; you should NOT change it.
    """

    def __init__(self, gameState, costFn = lambda x: 1, goal=(1,1), start=None, warn=True, visualize=True):
        """
        Stores the start and goal.

        gameState: A GameState object (pacman.py)
        costFn: A function from a search state (tuple) to a non-negative number
        goal: A position in the gameState
        """
        self.walls = gameState.getWalls()
        self.startState = gameState.getPacmanPosition()
        if start != None: self.startState = start
        self.goal = goal
        self.costFn = costFn
        self.visualize = visualize
        if warn and (gameState.getNumFood() != 1 or not gameState.hasFood(*goal)):
            print 'Warning: this does not look like a regular search maze'

        # For display purposes
        self._visited, self._visitedlist, self._expanded = {}, [], 0

    def getStartState(self):
        return self.startState

    def isGoalState(self, state):
        isGoal = state == self.goal

        # For display purposes only
        if isGoal and self.visualize:
            self._visitedlist.append(state)
            import __main__
            if '_display' in dir(__main__):
                if 'drawExpandedCells' in dir(__main__._display): #@UndefinedVariable
                    __main__._display.drawExpandedCells(self._visitedlist) #@UndefinedVariable

        return isGoal

    def getSuccessors(self, state):
        """
        Returns successor states, the actions they require, and a cost of 1.

         As noted in search.py:
             For a given state, this should return a list of triples,
         (successor, action, stepCost), where 'successor' is a
         successor to the current state, 'action' is the action
         required to get there, and 'stepCost' is the incremental
         cost of expanding to that successor
        """

        successors = []
        for action in [Directions.NORTH, Directions.SOUTH, Directions.EAST, Directions.WEST]:
            x,y = state
            dx, dy = Actions.directionToVector(action)
            nextx, nexty = int(x + dx), int(y + dy)
            if not self.walls[nextx][nexty]:
                nextState = (nextx, nexty)
                cost = self.costFn(nextState)
                successors.append( ( nextState, action, cost) )

        # Bookkeeping for display purposes
        self._expanded += 1
        if state not in self._visited:
            self._visited[state] = True
            self._visitedlist.append(state)

        return successors

    def getCostOfActions(self, actions):
        """
        Returns the cost of a particular sequence of actions.  If those actions
        include an illegal move, return 999999
        """
        if actions == None: return 999999
        x,y= self.getStartState()
        cost = 0
        for action in actions:
            # Check figure out the next state and see whether its' legal
            dx, dy = Actions.directionToVector(action)
            x, y = int(x + dx), int(y + dy)
            if self.walls[x][y]: return 999999
            cost += self.costFn((x,y))
        return cost

def manhattanHeuristic(position, problem, info={}):
    "The Manhattan distance heuristic for a PositionSearchProblem"
    xy1 = position
    xy2 = problem.goal
    return abs(xy1[0] - xy2[0]) + abs(xy1[1] - xy2[1])

def euclideanHeuristic(position, problem, info={}):
    "The Euclidean distance heuristic for a PositionSearchProblem"
    xy1 = position
    xy2 = problem.goal
    return ( (xy1[0] - xy2[0]) ** 2 + (xy1[1] - xy2[1]) ** 2 ) ** 0.5



#####################################################
#        Corners problem code, do not modify.       #
#####################################################

BOTTOM_LEFT = 0
TOP_LEFT = 1
BOTTOM_RIGHT = 2
TOP_RIGHT = 3

class CornerState:
   def __init__(self, startingPosition):
     self.currentPos = startingPosition
     self.cornersVisited = [0, 0, 0, 0]

   def __eq__(self, other):
     return self.currentPos == other.currentPos and self.cornersVisited == other.cornersVisited

   def __hash__(self):
     x = hash(self.currentPos)
     y = self.cornersVisited[BOTTOM_LEFT]*7 + self.cornersVisited[BOTTOM_LEFT]*29 + self.cornersVisited[BOTTOM_LEFT]*31 + self.cornersVisited[BOTTOM_LEFT]*37
     return hash(x + y)

   def __str__(self):
     return str(self.currentPos) + " " + str(self.cornersVisited)

   def getCurrentPos(self):
       """
       Returns the current position of Pacman as a tuple (x,y).
       """
       return self.currentPos

   def getCornersVisited(self):
       """
       Returns the corners visited as a list with a length of 4.
       
       Each element in the list indicates whether that corner has been visited.
       The corners are, in order:
       - Bottom left
       - Top left
       - Bottom right
       - Top right

       The order corresponds with the order of the problem.corners positions.
       """
       return self.cornersVisited

class CornersProblem(SearchProblem):
  """
  This search problem finds paths through all four corners of a layout.

  You must select a suitable state space and successor function
  """

  def __init__(self, startingGameState):
    """
    Stores the walls, pacman's starting position and corners.
    """
    self.walls = startingGameState.getWalls()
    self.startingPosition = startingGameState.getPacmanPosition()
    top, right = self.walls.height-2, self.walls.width-2 
    self.corners = ((1,1), (1,top), (right, 1), (right, top))
    for corner in self.corners:
      if not startingGameState.hasFood(*corner):
        print 'Warning: no food in corner ' + str(corner)
    self._expanded = 0 # Number of search nodes expanded
    self.heuristicInfo = {} # A dictionary for the heuristic to store information
    
  def getWalls(self):
      """
      Returns the walls of the maze as a grid.

      You can access wall areas by index as:
      getWalls()[x][y]
      where x denotes the number of squares from the left,
      and y denotes the number of squares from the bottom.

      For example, you can test whether the bottom left square contains a wall with:
      if getWalls()[0][0]:

      You can also get the size of the maze with: 
      getWalls().width
      and
      getWalls().height

      Note: in all mazes the outer ring is always true.
      """
      return self.walls

  def getCorners(self):
      """
      Returns the positions of all four corners as a list of (x,y) tuples.

       The corners are, in order:
       - Bottom left
       - Top left
       - Bottom right
       - Top right
      Note that this order corresponds with the order returned by state.getCornersVisited().
      """
      return self.corners

  def getHeuristicInfo(self):
      """
      Returns a dictionary in which you store arbitrary data to help your heuristic. 
      """
      return self.heuristicInfo
  
  def setHeuristicInfo(self, heuristicInfo):
      """
      Sets the heuristicInfo dictionary to be the dictionary provided.
      """
      self.heuristicInfo = heuristicInfo

  def getStartState(self):
    "Returns the start state (in your state space, not the full Pacman state space)"
    return CornerState(self.startingPosition)
    
  def isGoalState(self, state):
    "Returns whether this search state is a goal state of the problem"
    return sum(state.cornersVisited) == 4
       
  def getSuccessors(self, state):
    """
    Returns successor states, the actions they require, and a cost of 1.
    
     As noted in search.py:
         For a given state, this should return a list of triples, 
     (successor, action, stepCost), where 'successor' is a 
     successor to the current state, 'action' is the action
     required to get there, and 'stepCost' is the incremental 
     cost of expanding to that successor
    """
    successors = []
    
  
    for action in [Directions.NORTH, Directions.SOUTH, Directions.EAST, Directions.WEST]:
      # Add a successor state to the successor list if the action is legal
      # Here's a code snippet for figuring out whether a new position hits a wall:
      x,y = state.currentPos
      dx, dy = Actions.directionToVector(action)
      nextx, nexty = int(x + dx), int(y + dy)
      hitsWall = self.walls[nextx][nexty]
      if not(hitsWall):
        newstate = copy.deepcopy(state)
        newstate.currentPos = (nextx, nexty)
        if (nextx, nexty) == self.corners[BOTTOM_LEFT]:
          newstate.cornersVisited[BOTTOM_LEFT] = 1
        if (nextx, nexty) == self.corners[BOTTOM_RIGHT]:
          newstate.cornersVisited[BOTTOM_RIGHT] = 1
        if (nextx, nexty) == self.corners[TOP_LEFT]:
          newstate.cornersVisited[TOP_LEFT] = 1
        if (nextx, nexty) == self.corners[TOP_RIGHT]:
          newstate.cornersVisited[TOP_RIGHT] = 1
        successors.append((newstate, action, 1))
    self._expanded += 1

    return successors

  def getCostOfActions(self, actions):
      """
      Returns the cost of a particular sequence of actions.  If those actions
      include an illegal move, return 999999.  This is implemented for you.
      """
      if actions == None: return 999999
      x,y= self.startingPosition
      for action in actions:
          dx, dy = Actions.directionToVector(action)
          x, y = int(x + dx), int(y + dy)
          if self.walls[x][y]: return 999999
      return len(actions)


#####################################################
#         Food problem code, do not modify.        #
#####################################################

class FoodSearchProblem:
    """
    A search problem associated with finding the a path that collects all of the
    food (dots) in a Pacman game.

    A search state in this problem is a tuple ( pacmanPosition, foodGrid ) where
      pacmanPosition: a tuple (x,y) of integers specifying Pacman's position
      foodGrid:       a Grid (see game.py) of either True or False, specifying remaining food
    """
    def __init__(self, startingGameState):
        self.start = (startingGameState.getPacmanPosition(), startingGameState.getFood())
        self.walls = startingGameState.getWalls()
        self.startingGameState = startingGameState
        self._expanded = 0
        self.heuristicInfo = {} # A dictionary for the heuristic to store information

    def getWalls(self):
        """
        Returns the walls of the maze as a grid.

        You can access wall areas by index as:
        getWalls()[x][y]
        where x denotes the number of squares from the left,
        and y denotes the number of squares from the bottom.

        For example, you can test whether the bottom left square contains a wall with:
        if getWalls()[0][0]:

        You can also get the size of the maze with: 
        getWalls().width
        and
        getWalls().height

        Note: in all mazes the outer ring is always true.
        """
        return self.walls

    def getHeuristicInfo(self):
        """
        Returns a dictionary in which you store arbitrary data to help your heuristic. 
        """
        return self.heuristicInfo
  
    def setHeuristicInfo(self, heuristicInfo):
        """
        Sets the heuristicInfo dictionary to be the dictionary provided.
        """
        self.heuristicInfo = heuristicInfo

    def getStartState(self):
        """
        Returns the start state as a tuple of a Pacman Position and a Food Grid.
        """
        return self.start

    def isGoalState(self, state):
        """
        Checks whether the provided state is a goal state.
        """
        return state[1].count() == 0

    def getSuccessors(self, state):
        "Returns successor states, the actions they require, and a cost of 1."
        successors = []
        self._expanded += 1
        for direction in [Directions.NORTH, Directions.SOUTH, Directions.EAST, Directions.WEST]:
            x,y = state[0]
            dx, dy = Actions.directionToVector(direction)
            nextx, nexty = int(x + dx), int(y + dy)
            if not self.walls[nextx][nexty]:
                nextFood = state[1].copy()
                nextFood[nextx][nexty] = False
                successors.append( ( ((nextx, nexty), nextFood), direction, 1) )
        return successors

    def getCostOfActions(self, actions):
        """Returns the cost of a particular sequence of actions.  If those actions
        include an illegal move, return 999999"""
        x,y= self.getStartState()[0]
        cost = 0
        for action in actions:
            # figure out the next state and see whether it's legal
            dx, dy = Actions.directionToVector(action)
            x, y = int(x + dx), int(y + dy)
            if self.walls[x][y]:
                return 999999
            cost += 1
        return cost

#####################################################
# Any food problem code. Can be used for question 7 #
# Note that it currently relies on your breadth-    #
# first search to work.                             #
#####################################################
  
class AnyFoodSearchProblem(PositionSearchProblem):
  """
    A search problem for finding a path to any food.
    
    This search problem is just like the PositionSearchProblem, but
    has a different goal test. The state space and successor function 
    do not need to be changed.
    
    The class definition above, AnyFoodSearchProblem(PositionSearchProblem),
    inherits the methods of the PositionSearchProblem.
    
    You can use this search problem to help you fill in 
    the findPathToClosestDot method.
  """

  def __init__(self, gameState):
    "Stores information from the gameState.  You don't need to change this."
    # Store the food for later reference
    self.food = gameState.getFood()

    # Store info for the PositionSearchProblem (no need to change this)
    self.walls = gameState.getWalls()
    self.startState = gameState.getPacmanPosition()
    self.costFn = lambda x: 1
    self._visited, self._visitedlist, self._expanded = {}, [], 0
    
  def isGoalState(self, state):
    """
    The state is Pacman's position.
    """
    x,y = state
    return self.food[x][y]
